# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: http://fledge-iot.readthedocs.io/
# FOGLAMP_END

""" Bucket layer python client
"""

__author__ = "Amandeep Singh Arora"
__copyright__ = "Copyright (c) 2022 Dianomic Inc"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import aiohttp
import async_timeout
import http.client
import json
import logging
from abc import ABC, abstractmethod

from foglamp.common import logger
from foglamp.common.service_record import ServiceRecord
from foglamp.common.bucket_client.exceptions import *

_LOGGER = logger.setup(__name__, level=logging.INFO)


class AbstractBucket(ABC):
    """ abstract class for bucket client """

    def __init__(self):
        super().__init__()

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def disconnect(self):
        pass

    # Allow with context
    def __enter__(self):
        return self.connect()

    def __exit__(self, *args):
        self.disconnect()


class BucketClient(AbstractBucket):
    def __init__(self, core_management_host, core_management_port, svc=None):
        try:
            if svc:
                self.service = svc
            else:
                self.connect(core_management_host, core_management_port)

            self.base_url = '{}:{}'.format(self.service._address, self.service._port)
            self.management_api_url = '{}:{}'.format(self.service._address, self.service._management_port)
        except Exception:
            raise InvalidServiceInstance

    @property
    def base_url(self):
        return self.__base_url

    @base_url.setter
    def base_url(self, url):
        self.__base_url = url

    @property
    def service(self):
        return self.__service

    @service.setter
    def service(self, svc):
        if not isinstance(svc, ServiceRecord):
            w_msg = 'BucketService should be a valid Foglamp micro-service instance'
            _LOGGER.warning(w_msg)
            raise InvalidServiceInstance

        if not getattr(svc, "_type") == "BucketStorage":
            w_msg = 'BucketService should be a valid *Foglamp* micro-service instance'
            _LOGGER.warning(w_msg)
            raise InvalidServiceInstance

        self.__service = svc

    def _get_bucket_service(self, host, port):
        """ Get bucket service details to connect to it
            Args:
                host: Management client host
                port: Management client port
            Returns:
            Raises:
        """
        conn = http.client.HTTPConnection("{0}:{1}".format(host, port))
        # TODO: need to set http / https based on service protocol
        conn.request('GET', url='/foglamp/service?type=BucketStorage')
        r = conn.getresponse()
        if r.status in range(400, 500):
            _LOGGER.error("Get Service: Client error code: %d, %s", r.status, r.reason)
        if r.status in range(500, 600):
            _LOGGER.error("Get Service: Server error code: %d, %s", r.status, r.reason)

        res = r.read().decode()
        conn.close()
        response = json.loads(res)
        _LOGGER.info("_get_bucket_service: response={}".format(response))

        svc = response["services"][0]

        _LOGGER.info("_get_bucket_service: response: svc={}".format(svc))

        return svc

    def connect(self, core_management_host, core_management_port):
        """ Get the model files from the bucket service
            Args:
                core_management_host: Management client host
                core_management_port: Management client port
            Returns:
            Raises:
        """
        svc = self._get_bucket_service(host=core_management_host, port=core_management_port)
        _LOGGER.info("connect: len(svc)={}, svc={}".format(len(svc), svc))
        if len(svc) == 0:
            raise InvalidServiceInstance
        self.service = ServiceRecord(s_id=svc["id"], s_name=svc["name"], s_type=svc["type"], s_port=svc["service_port"],
                                     m_port=svc["management_port"], s_address=svc["address"],
                                     s_protocol=svc["protocol"])

        return self

    def disconnect(self):
        pass

    async def query_models(self, query):
        """ Use the bucket service API to get list of buckets with matching attribute values
        Args:
            query: List of key/value tuples to use for attribute based bucket search
        Returns:
            a JSON object containing information about matching buckets
        """
        put_url = '/bucket/match'
        url = 'http://' + self.base_url + put_url

        async with aiohttp.ClientSession() as session:
            async with session.put(url, data=query) as resp:
                status_code = resp.status
                jdoc = await resp.json()
                if status_code not in range(200, 209):
                    _LOGGER.info("PUT %s, with query: %s", put_url, data)
                    _LOGGER.error("Error code: %d, reason: %s, details: %s", resp.status, resp.reason, jdoc)
                    raise BucketServerError(code=resp.status, reason=resp.reason, error=jdoc)

        _LOGGER.info("query_models: result={}".format(jdoc))

        return jdoc

    async def get_bucket_content(self, bucket_id, filepath):
        """ Get bucket content
            Args:
                bucket_id: Bucket ID
                filepath: Filepath where to retrieve bucket contents
            Returns:
            Raises:
        """
        _TIME_OUT = 120
        _CHUNK_SIZE = 32*1024

        get_url = '/bucket/{}'.format(bucket_id)
        url = 'http://' + self.base_url + get_url

        _LOGGER.info("get_bucket_content: bucket_id={}, url={}".format(bucket_id, url))

        async with aiohttp.ClientSession() as session:
            async with async_timeout.timeout(_TIME_OUT):
                async with session.get(url) as resp:
                    status_code = resp.status
                    with open(filepath, 'wb') as fd:
                        async for data in resp.content.iter_chunked(_CHUNK_SIZE):
                            fd.write(data)

                    if status_code not in range(200, 209):
                        _LOGGER.info("GET %s", url)
                        _LOGGER.error("Error code: %d, reason: %s, details: %s", resp.status, resp.reason, "Error while getting bucket contents")
                        raise BucketServerError(code=resp.status, reason=resp.reason, error="Error while getting bucket contents")

        _LOGGER.info("get_bucket_contents: bucket_id={}, filepath={}, url={}, status_code={}"
                        .format(bucket_id, filepath, url, status_code))

        return
